from typing import Any

import grpc
import logging

from pyrdms.core.entities import SymbolType
from pyrdms.proto import rdm_pb2_grpc, rdm_pb2
from pyrdms.controllers.mappers import Py2Proto, Proto2Py

class RDMClient:
    def __init__(self, host: str = "localhost:50051") -> None:
        self.host = host

    def __enter__(self):
        self._chan = grpc.insecure_channel(self.host)
        self._stub = rdm_pb2_grpc.RemoteModelServiceStub(self._chan)
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        # if exc_val:
        #     logging.info(f"RDMClient exiting with error: {exc_val.debug_error_string()}")
        self._chan.close()

    def list(self) -> list:
        response: rdm_pb2.ListSignaturesResponse = self._stub.ListSignatures(rdm_pb2.ListSignaturesRequest())
        return response.libraries

    def __call__(self, type: SymbolType, model_name: str, function_name: str, *args: Any, **kwds: Any) -> Any:
        _args = [
            Py2Proto.map(arg)
            for arg in args
        ] + [
            Py2Proto.map(arg)
            for _, arg in kwds.items()
        ]
        
        response: rdm_pb2.CallResult = self._stub.Call(rdm_pb2.CallRequest(
            lib=rdm_pb2.Signature(name=model_name),
            type=rdm_pb2.FUNCTION if type != SymbolType.PREDICATE else rdm_pb2.PREDICATE,
            name=function_name,
            args=_args,
        ))

        return Proto2Py.map(response.value)


import logging
import grpc

from pyrdms.core.entities import SymbolType

from .app import serve
from pyrdms.client import RDMClient

import argparse
import os

from re import compile, findall

callstrre = compile(r"^([^.\ ]*)\.([^:\ ]*)\:(([^,\ ]+\,)*([^,\ ]*))$")

class HealthServer:
    def health(self) -> bool:
        logging.warning("Health request!")
        return True

class TestRDM:
    def test(self, a: bool, b: bool) -> bool:
        if a == None or b == None:
            return None
        return a and b

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    argsp = argparse.ArgumentParser()

    argsp.add_argument("-a", "--host", type=str, required=True, help="Hostname")
    argsp.add_argument("-p", "--port", type=int, required=False, default=50051, help="Port")

    argsp.add_argument("-c", "--client", action="store_true", help="Work as client")
    argsp.add_argument("-s", "--serve", action="store_true", help="Work as server")

    argsp.add_argument("--call-function", action="store_true", help="Add this flag if you want to call function with --call")
    argsp.add_argument("--call", type=str, help="Client argument: call string in format <model_name>.<predicate_name>:<arg1>,<arg2>,...")
    argsp.add_argument("--list", action="store_true", help="Client argument: list models")

    args = argsp.parse_args()

    if args.serve:
        hs = HealthServer()
        serve(port=args.port, health=hs, TestRDM=TestRDM())
    elif args.client:
        if args.list:
            with RDMClient(f"{args.host}:{args.port}") as cli:
                logging.info(f"Domain models on remote DM server: {cli.list()}")
            os._exit(0)

        if not args.call:
            print("call must not be nil!")
            os._exit(1)
        
        if not callstrre.match(args.call):
            logging.error(f"Invalid call string: {args.call}")
            os._exit(1)
        
        model, predicate, argsstr = callstrre.findall(args.call)[0][:3]

        callargs = []
        for arg in argsstr.split(','):
            if arg == "":
                continue

            if arg in ["True", "False", "true", "false"]:
                callargs.append(arg in ["True", "true"])
            else:
                try:
                    callargs.append(float(arg))
                except ValueError:
                    callargs.append(arg)

        try:
            with RDMClient(f"{args.host}:{args.port}") as cli:
                res = cli(SymbolType.FUNCTION if args.call_function else SymbolType.PREDICATE, model, predicate, *callargs)
            
            logging.info(f"Call result: {res}")
        except grpc.RpcError as ex:
            logging.warning(f"Got rpc error calling function: {ex}")
    else:
        argsp.print_help()
        os._exit(1)


from typing import List, Mapping
from dataclasses import dataclass
from enum import Enum

class SymbolType(Enum):
    FUNCTION = 1
    PREDICATE = 2

@dataclass
class DomainFunctionMeta:
    name: str
    args: List[type]
    ret: type

@dataclass
class DomainModelMeta:
    name: str
    domain_functions: Mapping[str, DomainFunctionMeta]



from .mappers import Proto2Py, Py2Proto

import json

from pyrdms.proto import rdm_pb2_grpc as rdm_pb2_grpc
from pyrdms.proto import rdm_pb2 as rdm_pb2

from pyrdms.core.entities import DomainModelMeta, DomainFunctionMeta

def test_map_meta():
    meta = DomainModelMeta(
        name="some_domain_model",
        domain_functions={
            "test1": DomainFunctionMeta(
                name="test1",
                args=[bool,int,str,list],
                ret=bool,
            )
        }
    )

    proto = Py2Proto.map_meta(meta)

    assert proto.name == meta.name
    assert proto.descriptors["test1"].name == "test1"
    assert proto.descriptors["test1"].return_value == rdm_pb2.LOGICAL
    assert len(proto.descriptors["test1"].arguments) == 4
    assert proto.descriptors["test1"].arguments[0] == rdm_pb2.LOGICAL
    assert proto.descriptors["test1"].arguments[1] == rdm_pb2.NUMERICAL
    assert proto.descriptors["test1"].arguments[2] == rdm_pb2.STRING
    assert proto.descriptors["test1"].arguments[3] == rdm_pb2.LIST

def test_map_int():
    val = Py2Proto.map(1)
    assert val.type == rdm_pb2.NUMERICAL
    assert val.value.numerical_value == 1.0

    assert Proto2Py.map(val) == 1.0

def test_map_float():
    val = Py2Proto.map(1.23)
    assert val.type == rdm_pb2.NUMERICAL
    assert abs(val.value.numerical_value - 1.23) < 1e-9

    assert abs(Proto2Py.map(val) - 1.23) < 1e-9

def test_map_str():
    val = Py2Proto.map("123")
    assert val.type == rdm_pb2.STRING
    assert val.value.string_value == "123"

    assert Proto2Py.map(val) == "123"

def test_map_bool():
    val1 = Py2Proto.map(None)
    val2 = Py2Proto.map(True)

    assert val1.type == rdm_pb2.LOGICAL
    assert val1.value.logical_value == rdm_pb2.NONE

    assert val2.type == rdm_pb2.LOGICAL
    assert val2.value.logical_value == rdm_pb2.TRUE

    assert Proto2Py.map(val1) == None
    assert Proto2Py.map(val2) == True

def test_map_list():
    some_list = [1, True, "123"]

    val = Py2Proto.map(some_list)

    assert val.type == rdm_pb2.LIST
    assert len(val.list_of_values.data) == 3

    assert val.list_of_values.data[0].type == rdm_pb2.NUMERICAL
    assert val.list_of_values.data[1].type == rdm_pb2.LOGICAL
    assert val.list_of_values.data[2].type == rdm_pb2.STRING

    assert val.list_of_values.data[0].value.numerical_value == 1.0
    assert val.list_of_values.data[1].value.logical_value == rdm_pb2.TRUE
    assert val.list_of_values.data[2].value.string_value == "123" 

    assert Proto2Py.map(val) == some_list

def test_map_list_of_lists():
    some_list = [1, True, "123", [1, True, "123"]]

    val = Py2Proto.map(some_list)

    assert val.type == rdm_pb2.LIST
    assert len(val.list_of_values.data) == 4

    assert val.list_of_values.data[0].type == rdm_pb2.NUMERICAL
    assert val.list_of_values.data[1].type == rdm_pb2.LOGICAL
    assert val.list_of_values.data[2].type == rdm_pb2.STRING

    assert val.list_of_values.data[0].value.numerical_value == 1.0
    assert val.list_of_values.data[1].value.logical_value == rdm_pb2.TRUE
    assert val.list_of_values.data[2].value.string_value == "123" 

    assert val.list_of_values.data[3].type == rdm_pb2.LIST
    assert len(val.list_of_values.data[3].list_of_values.data) == 3

    assert val.list_of_values.data[3].list_of_values.data[0].type == rdm_pb2.NUMERICAL
    assert val.list_of_values.data[3].list_of_values.data[1].type == rdm_pb2.LOGICAL
    assert val.list_of_values.data[3].list_of_values.data[2].type == rdm_pb2.STRING

    assert val.list_of_values.data[3].list_of_values.data[0].value.numerical_value == 1.0
    assert val.list_of_values.data[3].list_of_values.data[1].value.logical_value == rdm_pb2.TRUE
    assert val.list_of_values.data[3].list_of_values.data[2].value.string_value == "123" 

    assert Proto2Py.map(val) == some_list

def test_map_dict():
    data = json.loads("""{
  "basicTypes": {
    "string": "Hello, World!",
    "number": 42,
    "float": 3.14159,
    "negative": -7.5,
    "booleanTrue": true,
    "booleanFalse": false,
    "nullValue": null
  },
  "unicodeStrings": {
    "emoji": "Hello 🌍 World! 🚀",
    "mixedUnicode": "中文 Español Français العربية"
  },
  "edgeCases": {
    "emptyString": "",
    "zero": 0,
    "negativeZero": -0,
    "largeNumber": 1.23456789e+30,
    "smallNumber": 1.23456789e-30,
    "maxSafeInteger": 9007199254740991,
    "minSafeInteger": -9007199254740991
  },
  "complexStructures": {
    "nestedObject": {
      "level1": {
        "level2": {
          "level3": "deep value",
          "array": [1, 2, 3]
        }
      }
    },
    "mixedArray": [
      "string",
      42,
      true,
      false,
      null,
      {"object": "in array"},
      ["nested", "array"]
    ],
    "emptyArrays": {
      "emptyArray": [],
      "emptyObject": {}
    }
  },
  "specialKeys": {
    "": "empty key",
    "key with spaces": "value",
    "key-with-dashes": "value",
    "key_with_underscores": "value",
    "keyWithCamelCase": "value",
    "KeyWithPascalCase": "value",
    "123numeric": "numeric first",
    "special!@#$%^&*()": "special chars in key"
  },
  "boundaryValues": {
    "maxDecimals": 0.1234567890123456,
    "scientificNotation": [1e10, 1e-10, -1e10, -1e-10],
    "infinityLike": 1.7976931348623157e+308
  },
  "dateTimeFormats": {
    "isoDate": "2023-12-25T10:30:00.000Z",
    "dateOnly": "2023-12-25",
    "timeOnly": "10:30:00",
    "withTimezone": "2023-12-25T10:30:00+01:00"
  },
  "escapeSequences": {
    "backslashes": "\\\\backslashes\\\\",
    "unicodeEscape": "\\u00A9 Copyright \\u2764"
  },
  "deepStructure": {
    "very": {
      "deeply": {
        "nested": {
          "object": {
            "with": {
              "many": {
                "levels": {
                  "of": {
                    "nesting": "final value",
                    "array": [
                      {
                        "nestedInArray": {
                          "anotherLevel": "array value"
                        }
                      }
                    ]
                  }
                }
              }
            }
          }
        }
      }
    }
  },
  "largeArray": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20],
  "booleanArray": [true, false, true, false],
  "nullArray": [null, null, null],
  "mixedEmpty": [null, "", 0, false, [], {}]
}""")
    
    assert Proto2Py.map(Py2Proto.map(data)) == data


from abc import ABC, ABCMeta, abstractmethod
from typing import Any, List

from pyrdms.core.entities import DomainFunctionMeta, DomainModelMeta

class DomainModelRepo(ABC):
    @abstractmethod
    def register_model(name: str, model: Any):
        raise NotImplementedError()

    @abstractmethod
    def __call__(self, model_name: str, predicate_name: str, *args: Any, **kwds: Any) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def has(self, model_name: str, predicate: DomainFunctionMeta) -> bool:
        raise NotImplementedError()
    
    @abstractmethod
    def get_models(self) -> List[DomainModelMeta]:
        raise NotImplementedError()


from .local import ClassDomainModelRepo
from .util import get_decorators
from .decorators import predicate



from typing import Union

from pyrdms.repo import DomainModelRepo

import logging

class RDMServer:
    def __init__(self, repo: DomainModelRepo) -> None:
        self._repo = repo

    def call(self, domain_library: str, symbol_name: str, args: list) -> Union[int, float, str, bool, list, dict]:
        logging.info(f"Calling predicate {domain_library}.{symbol_name}({args})")

        return self._repo(domain_library, symbol_name, *args)
    
    def list_libraries(self) -> map:
        models = self._repo.get_models()

        logging.info(f"Models: {models}")

        return models



from . import ClassDomainModelRepo
from pyrdms.repo.decorators import predicate

from pyrdms.core.entities import DomainFunctionMeta, DomainModelMeta

class SomeClass:
    def __init__(self, a, b, c):
        pass

    def some_method(self, a: int, b: str, c: float, d: list, e: bool) -> int:
        return 123

class SomeClassDecorated:
    def __init__(self, a, b, c):
        pass

    def some_not_decorated_method(self, a: int, b: str, c: float, d: list, e: bool) -> int:
        return 123

    @predicate
    def some_method(self, a: int, b: str, c: float, d: list, e: bool) -> int:
        return 321

def test_local_domain_model():
    repo = ClassDomainModelRepo()
    cl = SomeClass(1,2,3)

    repo.register_model("test", cl)

    assert repo("test", "some_method", 1, "2", 3.0, [], True) == 123
    assert repo.has("test", predicate=DomainFunctionMeta(
        name="some_method",
        args=[int, str, float, list, bool],
        ret=int,
    ))

    assert repo.get_models() == [DomainModelMeta(
        name="test",
        domain_functions={"some_method": DomainFunctionMeta(
            name="some_method",
            args=[int, str, float, list, bool],
            ret=int,
        )}
    )]

    # assert print(repo._dms) != None

def test_local_domain_model_decorated():
    repo = ClassDomainModelRepo()
    cl = SomeClassDecorated(1,2,3)

    repo.register_model("test", cl)

    assert repo("test", "some_method", 1, "2", 3.0, [], True) == 321
    assert repo.has("test", predicate=DomainFunctionMeta(
        name="some_method",
        args=[int, str, float, list, bool],
        ret=int,
    ))

    assert not repo.has("test", predicate=DomainFunctionMeta(
        name="some_not_decorated_method",
        args=[int, str, float, list, bool],
        ret=int,
    ))

    assert repo.get_models() == [DomainModelMeta(
        name="test",
        domain_functions={"some_method": DomainFunctionMeta(
            name="some_method",
            args=[int, str, float, list, bool],
            ret=int,
        )}
    )]

    # assert print(repo._dms) != None



# def predicate(func):
#     """
#     predicate decorator just marks that method of domain model must be exposed 
#     """
#     def wrapper(*args, **kwargs):
#         return func(*args, **kwargs)
#     return wrapper

class predicate:
    def __init__(self, function):
        self.function = function
     
    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)


from . import DomainModelRepo
from pyrdms.repo.decorators import predicate
from pyrdms.repo.util import get_decorators

import inspect
from typing import Any, Mapping

from pyrdms.core.entities import DomainFunctionMeta, DomainModelMeta
from pyrdms.exceptions import PredicateNotFoundException

class ClassDomainModelRepo(DomainModelRepo):
    def __init__(self) -> None:
        super().__init__()
        self._dms = {}
        self._signs = {}
        self._models = {}
    
    @staticmethod
    def _get_model_decorated_methods(model: object):
        decorators = get_decorators(model.__class__)
        return [
            method_name
            for method_name, decorators in decorators.items() if 'predicate' in decorators
        ]
    
    @staticmethod
    def _method_search_predicate(object: object):
        """Returns if object is method or if object is method decorated with predicate"""
        return inspect.ismethod(object) or isinstance(object, predicate)
    
    def register_model(self, name: str, model):
        # Проверяем, что нам передали экземпляр класса
        assert not isinstance(model, type)

        # todo: тут надо больше проверок, чтобы не выстрелить в ногу
        self._dms[name] = {}
        self._signs[name] = {}

        decorated_methods = self._get_model_decorated_methods(model)

        for method_name, method in inspect.getmembers(model, self._method_search_predicate):
            print(f"Check {method_name}")
            # Нас не интересуют магические методы и, при наличии декораторов, интересуют только декорированные с помощью 'predicate'
            if not method_name.startswith("__") and (not decorated_methods or method_name in decorated_methods):
                self._dms[name][method_name] = method

                if method_name not in decorated_methods:
                    sign = inspect.signature(method)
                else:
                    # get underlying function of predicate decorator
                    sign = inspect.signature(method.function)

                    # model to use as self parameter to method
                    self._models[name] = model

                print(f"register method with name {method_name} and signature {sign} params: {list(sign.parameters.items())} of {model}")
                
                self._signs[name][method_name] = DomainFunctionMeta(
                    name=method_name,
                    # for some reason self is returned when method is decorated as a parameter here
                    args=[v.annotation for k, v in sign.parameters.items() if k != 'self'],
                    ret=sign.return_annotation,
                )
    
    def __call__(self, model_name: str, predicate_name: str, *args: Any, **kwds: Any) -> Any:
        if model_name not in self._dms:
            raise PredicateNotFoundException(model_name)
        
        if predicate_name not in self._dms[model_name]:
            raise PredicateNotFoundException(model_name)
        
        if model_name not in self._models:
            # no need to pass self parameter, because method is already IS bound method of class
            return self._dms[model_name][predicate_name](*args, **kwds)

        # method is decorated in class, so we need to pass the model itself as first parameter 
        return self._dms[model_name][predicate_name](self._models[model_name], *args, **kwds)

    def has(self, model_name: str, predicate: DomainFunctionMeta) -> bool:
        return model_name in self._dms and \
            predicate.name in self._dms[model_name] and \
            self._signs[model_name][predicate.name] == predicate
    
    def get_models(self) -> Mapping[str, DomainModelMeta]:
        meta = []

        for model_name in self._dms.keys():
            meta.append(
                DomainModelMeta(
                    name=model_name,
                    domain_functions={
                        predicate: descr
                        for predicate, descr in self._signs[model_name].items()
                    },
                )
            )

        return meta



from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Logical(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FALSE: _ClassVar[Logical]
    TRUE: _ClassVar[Logical]
    NONE: _ClassVar[Logical]

class SemanticValueType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    STRING: _ClassVar[SemanticValueType]
    NUMERICAL: _ClassVar[SemanticValueType]
    LOGICAL: _ClassVar[SemanticValueType]
    LIST: _ClassVar[SemanticValueType]
    OBJECT: _ClassVar[SemanticValueType]

class SymbolType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PREDICATE: _ClassVar[SymbolType]
    FUNCTION: _ClassVar[SymbolType]
FALSE: Logical
TRUE: Logical
NONE: Logical
STRING: SemanticValueType
NUMERICAL: SemanticValueType
LOGICAL: SemanticValueType
LIST: SemanticValueType
OBJECT: SemanticValueType
PREDICATE: SymbolType
FUNCTION: SymbolType

class BaseSemanticValue(_message.Message):
    __slots__ = ("string_value", "numerical_value", "logical_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    NUMERICAL_VALUE_FIELD_NUMBER: _ClassVar[int]
    LOGICAL_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    numerical_value: float
    logical_value: Logical
    def __init__(self, string_value: _Optional[str] = ..., numerical_value: _Optional[float] = ..., logical_value: _Optional[_Union[Logical, str]] = ...) -> None: ...

class ListOfSemanticValues(_message.Message):
    __slots__ = ("data",)
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedCompositeFieldContainer[SemanticValue]
    def __init__(self, data: _Optional[_Iterable[_Union[SemanticValue, _Mapping]]] = ...) -> None: ...

class SemanticObject(_message.Message):
    __slots__ = ("iri", "data")
    IRI_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    iri: str
    data: _struct_pb2.Struct
    def __init__(self, iri: _Optional[str] = ..., data: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...

class SemanticValue(_message.Message):
    __slots__ = ("type", "value", "list_of_values", "object_value")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    LIST_OF_VALUES_FIELD_NUMBER: _ClassVar[int]
    OBJECT_VALUE_FIELD_NUMBER: _ClassVar[int]
    type: SemanticValueType
    value: BaseSemanticValue
    list_of_values: ListOfSemanticValues
    object_value: SemanticObject
    def __init__(self, type: _Optional[_Union[SemanticValueType, str]] = ..., value: _Optional[_Union[BaseSemanticValue, _Mapping]] = ..., list_of_values: _Optional[_Union[ListOfSemanticValues, _Mapping]] = ..., object_value: _Optional[_Union[SemanticObject, _Mapping]] = ...) -> None: ...

class SymbolDescriptor(_message.Message):
    __slots__ = ("type", "name", "arguments", "return_value")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ARGUMENTS_FIELD_NUMBER: _ClassVar[int]
    RETURN_VALUE_FIELD_NUMBER: _ClassVar[int]
    type: SymbolType
    name: str
    arguments: _containers.RepeatedScalarFieldContainer[SemanticValueType]
    return_value: SemanticValueType
    def __init__(self, type: _Optional[_Union[SymbolType, str]] = ..., name: _Optional[str] = ..., arguments: _Optional[_Iterable[_Union[SemanticValueType, str]]] = ..., return_value: _Optional[_Union[SemanticValueType, str]] = ...) -> None: ...

class Signature(_message.Message):
    __slots__ = ("fqn", "name", "descriptors")
    class DescriptorsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: SymbolDescriptor
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[SymbolDescriptor, _Mapping]] = ...) -> None: ...
    FQN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTORS_FIELD_NUMBER: _ClassVar[int]
    fqn: str
    name: str
    descriptors: _containers.MessageMap[str, SymbolDescriptor]
    def __init__(self, fqn: _Optional[str] = ..., name: _Optional[str] = ..., descriptors: _Optional[_Mapping[str, SymbolDescriptor]] = ...) -> None: ...

class Caller(_message.Message):
    __slots__ = ("modelName", "name")
    MODELNAME_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    modelName: str
    name: str
    def __init__(self, modelName: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class CalculationContext(_message.Message):
    __slots__ = ("caller",)
    CALLER_FIELD_NUMBER: _ClassVar[int]
    caller: Caller
    def __init__(self, caller: _Optional[_Union[Caller, _Mapping]] = ...) -> None: ...

class ListSignaturesRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListSignaturesResponse(_message.Message):
    __slots__ = ("libraries",)
    LIBRARIES_FIELD_NUMBER: _ClassVar[int]
    libraries: _containers.RepeatedCompositeFieldContainer[Signature]
    def __init__(self, libraries: _Optional[_Iterable[_Union[Signature, _Mapping]]] = ...) -> None: ...

class CallRequest(_message.Message):
    __slots__ = ("lib", "type", "name", "args", "ctx")
    LIB_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    CTX_FIELD_NUMBER: _ClassVar[int]
    lib: Signature
    type: SymbolType
    name: str
    args: _containers.RepeatedCompositeFieldContainer[SemanticValue]
    ctx: CalculationContext
    def __init__(self, lib: _Optional[_Union[Signature, _Mapping]] = ..., type: _Optional[_Union[SymbolType, str]] = ..., name: _Optional[str] = ..., args: _Optional[_Iterable[_Union[SemanticValue, _Mapping]]] = ..., ctx: _Optional[_Union[CalculationContext, _Mapping]] = ...) -> None: ...

class CallResult(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: SemanticValue
    def __init__(self, value: _Optional[_Union[SemanticValue, _Mapping]] = ...) -> None: ...



import ast
import inspect

def get_decorators(cls):
    """
    Example:
    ```python
def template(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper

baz = template
che = template

class Foo(object):
    @baz
    @che
    def bar(self):
        pass
    ```
    Out: `{'bar': ['baz', 'che']}`
    """

    target = cls
    decorators = {}

    def visit_FunctionDef(node: ast.FunctionDef):
        decorators[node.name] = []
    
        for n in node.decorator_list:
            name = ''
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            decorators[node.name].append(name)

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    node_iter.visit(ast.parse(inspect.getsource(target)))
    return decorators



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: rdm.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'rdm.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\trdm.proto\x12\x0corg.d0sl.rdm\x1a\x1cgoogle/protobuf/struct.proto\"\x7f\n\x11\x42\x61seSemanticValue\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x19\n\x0fnumerical_value\x18\x02 \x01(\x01H\x00\x12.\n\rlogical_value\x18\x03 \x01(\x0e\x32\x15.org.d0sl.rdm.LogicalH\x00\x42\x07\n\x05value\"A\n\x14ListOfSemanticValues\x12)\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x1b.org.d0sl.rdm.SemanticValue\"D\n\x0eSemanticObject\x12\x0b\n\x03iri\x18\x01 \x01(\t\x12%\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"\xeb\x01\n\rSemanticValue\x12-\n\x04type\x18\x01 \x01(\x0e\x32\x1f.org.d0sl.rdm.SemanticValueType\x12\x30\n\x05value\x18\x02 \x01(\x0b\x32\x1f.org.d0sl.rdm.BaseSemanticValueH\x00\x12<\n\x0elist_of_values\x18\x03 \x01(\x0b\x32\".org.d0sl.rdm.ListOfSemanticValuesH\x00\x12\x34\n\x0cobject_value\x18\x04 \x01(\x0b\x32\x1c.org.d0sl.rdm.SemanticObjectH\x00\x42\x05\n\x03val\"\xb3\x01\n\x10SymbolDescriptor\x12&\n\x04type\x18\x01 \x01(\x0e\x32\x18.org.d0sl.rdm.SymbolType\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x32\n\targuments\x18\x03 \x03(\x0e\x32\x1f.org.d0sl.rdm.SemanticValueType\x12\x35\n\x0creturn_value\x18\x04 \x01(\x0e\x32\x1f.org.d0sl.rdm.SemanticValueType\"\xb9\x01\n\tSignature\x12\x0b\n\x03\x66qn\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12=\n\x0b\x64\x65scriptors\x18\x03 \x03(\x0b\x32(.org.d0sl.rdm.Signature.DescriptorsEntry\x1aR\n\x10\x44\x65scriptorsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.org.d0sl.rdm.SymbolDescriptor:\x02\x38\x01\")\n\x06\x43\x61ller\x12\x11\n\tmodelName\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\":\n\x12\x43\x61lculationContext\x12$\n\x06\x63\x61ller\x18\x01 \x01(\x0b\x32\x14.org.d0sl.rdm.Caller\"\x17\n\x15ListSignaturesRequest\"D\n\x16ListSignaturesResponse\x12*\n\tlibraries\x18\x01 \x03(\x0b\x32\x17.org.d0sl.rdm.Signature\"\xd0\x01\n\x0b\x43\x61llRequest\x12$\n\x03lib\x18\x01 \x01(\x0b\x32\x17.org.d0sl.rdm.Signature\x12&\n\x04type\x18\x02 \x01(\x0e\x32\x18.org.d0sl.rdm.SymbolType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12)\n\x04\x61rgs\x18\x04 \x03(\x0b\x32\x1b.org.d0sl.rdm.SemanticValue\x12\x32\n\x03\x63tx\x18\x05 \x01(\x0b\x32 .org.d0sl.rdm.CalculationContextH\x00\x88\x01\x01\x42\x06\n\x04_ctx\"8\n\nCallResult\x12*\n\x05value\x18\x01 \x01(\x0b\x32\x1b.org.d0sl.rdm.SemanticValue*(\n\x07Logical\x12\t\n\x05\x46\x41LSE\x10\x00\x12\x08\n\x04TRUE\x10\x01\x12\x08\n\x04NONE\x10\x02*Q\n\x11SemanticValueType\x12\n\n\x06STRING\x10\x00\x12\r\n\tNUMERICAL\x10\x01\x12\x0b\n\x07LOGICAL\x10\x02\x12\x08\n\x04LIST\x10\x03\x12\n\n\x06OBJECT\x10\x04*)\n\nSymbolType\x12\r\n\tPREDICATE\x10\x00\x12\x0c\n\x08\x46UNCTION\x10\x01\x32\xae\x01\n\x12RemoteModelService\x12;\n\x04\x43\x61ll\x12\x19.org.d0sl.rdm.CallRequest\x1a\x18.org.d0sl.rdm.CallResult\x12[\n\x0eListSignatures\x12#.org.d0sl.rdm.ListSignaturesRequest\x1a$.org.d0sl.rdm.ListSignaturesResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rdm_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SIGNATURE_DESCRIPTORSENTRY']._loaded_options = None
  _globals['_SIGNATURE_DESCRIPTORSENTRY']._serialized_options = b'8\001'
  _globals['_LOGICAL']._serialized_start=1398
  _globals['_LOGICAL']._serialized_end=1438
  _globals['_SEMANTICVALUETYPE']._serialized_start=1440
  _globals['_SEMANTICVALUETYPE']._serialized_end=1521
  _globals['_SYMBOLTYPE']._serialized_start=1523
  _globals['_SYMBOLTYPE']._serialized_end=1564
  _globals['_BASESEMANTICVALUE']._serialized_start=57
  _globals['_BASESEMANTICVALUE']._serialized_end=184
  _globals['_LISTOFSEMANTICVALUES']._serialized_start=186
  _globals['_LISTOFSEMANTICVALUES']._serialized_end=251
  _globals['_SEMANTICOBJECT']._serialized_start=253
  _globals['_SEMANTICOBJECT']._serialized_end=321
  _globals['_SEMANTICVALUE']._serialized_start=324
  _globals['_SEMANTICVALUE']._serialized_end=559
  _globals['_SYMBOLDESCRIPTOR']._serialized_start=562
  _globals['_SYMBOLDESCRIPTOR']._serialized_end=741
  _globals['_SIGNATURE']._serialized_start=744
  _globals['_SIGNATURE']._serialized_end=929
  _globals['_SIGNATURE_DESCRIPTORSENTRY']._serialized_start=847
  _globals['_SIGNATURE_DESCRIPTORSENTRY']._serialized_end=929
  _globals['_CALLER']._serialized_start=931
  _globals['_CALLER']._serialized_end=972
  _globals['_CALCULATIONCONTEXT']._serialized_start=974
  _globals['_CALCULATIONCONTEXT']._serialized_end=1032
  _globals['_LISTSIGNATURESREQUEST']._serialized_start=1034
  _globals['_LISTSIGNATURESREQUEST']._serialized_end=1057
  _globals['_LISTSIGNATURESRESPONSE']._serialized_start=1059
  _globals['_LISTSIGNATURESRESPONSE']._serialized_end=1127
  _globals['_CALLREQUEST']._serialized_start=1130
  _globals['_CALLREQUEST']._serialized_end=1338
  _globals['_CALLRESULT']._serialized_start=1340
  _globals['_CALLRESULT']._serialized_end=1396
  _globals['_REMOTEMODELSERVICE']._serialized_start=1567
  _globals['_REMOTEMODELSERVICE']._serialized_end=1741
# @@protoc_insertion_point(module_scope)



from concurrent import futures
import grpc

from typing import Mapping, Any, Callable
from functools import partial

import logging

from pyrdms.controllers import RemoteDomainModelController
from pyrdms.core.services import RDMServer
from pyrdms.repo import ClassDomainModelRepo

def start_server(port: int, **models) -> grpc.Server:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

    repo = ClassDomainModelRepo()

    for name, model in models.items():
        repo.register_model(name, model)

    svc = RDMServer(repo)

    controller = RemoteDomainModelController(svc)

    controller.register_on_server(server)

    server.add_insecure_port(f'[::]:{port}')
    server.start()

    logging.warning(f"Server started, listening on {port}")

    return server

def serve(port: int, **models):
    server = start_server(port, **models)
    server.wait_for_termination()

def serve_background(port: int, **models) -> Callable[[], None]:
    server = start_server(port, **models)
    return partial(server.stop, grace=3)



import grpc
from pyrdms.proto import rdm_pb2_grpc as rdm_pb2_grpc
from pyrdms.proto import rdm_pb2 as rdm_pb2

from pyrdms.controllers.mappers import Proto2Py, Py2Proto

from pyrdms.core.services import RDMServer

import logging

class RemoteDomainModelController(rdm_pb2_grpc.RemoteModelService):
    def __init__(self, service: RDMServer) -> None:
        super().__init__()
        self.service = service

    def register_on_server(self, server):
        rdm_pb2_grpc.add_RemoteModelServiceServicer_to_server(self, server)

    def Call(self, request: rdm_pb2.CallRequest, context):
        logging.debug(f"CallPredicate: {request}")
        
        context.set_code(grpc.StatusCode.OK)
        context.set_details('ok')

        try:
            py_result = self.service.call(
                domain_library=request.lib.name,
                symbol_name=request.name,
                args=[
                    Proto2Py.map(e)
                    for e in request.args
                ],
            )

            result = Py2Proto.map(py_result)
        except Exception as e:
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details(f'Error: {e}')

            logging.error(f"Error occured while processing request: {request}", e)
            return rdm_pb2.CallResult(
                value=rdm_pb2.SemanticValue(
                    type=rdm_pb2.LOGICAL,
                    value=rdm_pb2.BaseSemanticValue(
                        logical_value=rdm_pb2.NONE,
                    )
                ))
        
        logging.info(f"Return result: {result}")
        
        return rdm_pb2.CallResult(value=result)
        
    
    def ListSignatures(self, request: rdm_pb2.ListSignaturesRequest, context):
        logging.debug(f"ListLibraries: {request}")
        context.set_code(grpc.StatusCode.OK)
        context.set_details('ok')

        try:
            py_result = self.service.list_libraries()
        except Exception as e:
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details(f'error: {e}')

            logging.error(f"Error occured while processing request: {request}", e)
            return rdm_pb2.ListSignaturesResponse(libraries=[])
        
        return rdm_pb2.ListSignaturesResponse(libraries=[
            Py2Proto.map_meta(lib)
            for lib in py_result
        ])



from typing import Union, List, Optional

from pyrdms.proto import rdm_pb2_grpc as rdm_pb2_grpc
from pyrdms.proto import rdm_pb2 as rdm_pb2

from pyrdms.core.entities import DomainModelMeta

import google.protobuf.struct_pb2 as struct_pb2
from typing import Any, Dict, List, Union

class Proto2Py:
    @staticmethod
    def __value_to_dict(value: struct_pb2.Value) -> Any:
        if value.HasField('null_value'):
            return None
        elif value.HasField('number_value'):
            return value.number_value
        elif value.HasField('string_value'):
            return value.string_value
        elif value.HasField('struct_value'):
            return Proto2Py.__struct_to_dict(value.struct_value)
        elif value.HasField('list_value'):
            return [Proto2Py.__value_to_dict(item) for item in value.list_value.values]
        else:
            raise ValueError("Unsupported Value type")

    @staticmethod
    def __struct_to_dict(struct: struct_pb2.Struct) -> Dict[str, Any]:
        return {
            key: Proto2Py.__value_to_dict(value)
            for key, value in struct.fields.items()
        }

    @staticmethod
    def _map_string(v: rdm_pb2.SemanticValue) -> str:
        assert v.type == rdm_pb2.STRING

        return v.value.string_value

    @staticmethod
    def _map_num(v: rdm_pb2.SemanticValue) -> float:
        assert v.type == rdm_pb2.NUMERICAL

        return v.value.numerical_value
    
    @staticmethod
    def _map_bool(v: rdm_pb2.SemanticValue) -> Optional[bool]:
        assert v.type == rdm_pb2.LOGICAL

        return None if v.value.logical_value == rdm_pb2.NONE else v.value.logical_value == rdm_pb2.TRUE
    
    @staticmethod
    def _map_list(l: rdm_pb2.SemanticValue) -> list:
        assert l.type == rdm_pb2.LIST

        return [Proto2Py.map(e) for e in l.list_of_values.data]

    @staticmethod
    def _map_object(o: rdm_pb2.SemanticValue) -> dict:
        return Proto2Py.__struct_to_dict(o.object_value.data)

    @staticmethod
    def map(value: rdm_pb2.SemanticValue) -> Union[list, int, float, str, Optional[bool], dict]:
        match value.type:
            case rdm_pb2.LOGICAL:
                return Proto2Py._map_bool(value)
            case rdm_pb2.NUMERICAL:
                return Proto2Py._map_num(value)
            case rdm_pb2.STRING:
                return Proto2Py._map_string(value)
            case rdm_pb2.LIST:
                return Proto2Py._map_list(value)
            case rdm_pb2.OBJECT:
                return Proto2Py._map_object(value)
            case _:
                raise TypeError(f"wrong value type: {type(value.type)}")


class Py2Proto:
    @staticmethod
    def __dict_to_value(value: Any) -> struct_pb2.Value:
        """
        Convert a Python value to a protobuf Value object.
        """
        if value is None:
            return struct_pb2.Value(null_value=struct_pb2.NullValue.NULL_VALUE)
        elif isinstance(value, (int, float)):
            return struct_pb2.Value(number_value=value)
        elif isinstance(value, str):
            return struct_pb2.Value(string_value=value)
        elif isinstance(value, dict):
            return struct_pb2.Value(struct_value=struct_pb2.Struct(fields={
                key: Py2Proto.__dict_to_value(val)
                for key, val in value.items()
            }))
        elif isinstance(value, list):
            return struct_pb2.Value(list_value=struct_pb2.ListValue(values=[
                Py2Proto.__dict_to_value(item)
                for item in value
            ]))
        else:
            raise ValueError(f"Unsupported type: {type(value)}")

    @staticmethod
    def __dict_to_struct(data: dict) -> struct_pb2.Struct:
        """
        Convert a Python dictionary to a protobuf Struct object.
        """
        return struct_pb2.Struct(
            fields={
                key: Py2Proto.__dict_to_value(val)
                for key, val in data.items()
            }
        )

    @staticmethod
    def _map_string(s: str) -> rdm_pb2.SemanticValue:
        assert isinstance(s, str)

        return rdm_pb2.SemanticValue(type=rdm_pb2.STRING, value=rdm_pb2.BaseSemanticValue(string_value=s))

    @staticmethod
    def _map_num(s: Union[int, float]) -> rdm_pb2.SemanticValue:
        assert isinstance(s, int) or isinstance(s, float)

        return rdm_pb2.SemanticValue(type=rdm_pb2.NUMERICAL, value=rdm_pb2.BaseSemanticValue(numerical_value=s))
    
    @staticmethod
    def _map_bool(b: bool) -> rdm_pb2.SemanticValue:
        assert b == None or isinstance(b, bool)

        if b == None:
            return rdm_pb2.SemanticValue(type=rdm_pb2.LOGICAL, value=rdm_pb2.BaseSemanticValue(logical_value=rdm_pb2.NONE))
        
        return rdm_pb2.SemanticValue(type=rdm_pb2.LOGICAL, value=rdm_pb2.BaseSemanticValue(logical_value=rdm_pb2.TRUE if b else rdm_pb2.FALSE))

    @staticmethod
    def _map_list(l: list | tuple) -> rdm_pb2.SemanticValue:
        return rdm_pb2.SemanticValue(
                type=rdm_pb2.LIST,
                list_of_values=rdm_pb2.ListOfSemanticValues(
                    data=[
                        Py2Proto.map(v)
                        for v in l
                    ]
                )
            )
    
    @staticmethod
    def _map_dict(value: dict) -> rdm_pb2.SemanticValue:
        return rdm_pb2.SemanticValue(
            type=rdm_pb2.OBJECT,
            object_value=rdm_pb2.SemanticObject(
                data=Py2Proto.__dict_to_struct(value),
            ),
        )

    @staticmethod
    def _map_type(t: type) -> rdm_pb2.SemanticValueType:
        if t == bool:
            return rdm_pb2.LOGICAL
        elif t == int or t == float:
            return rdm_pb2.NUMERICAL
        elif t == str:
            return rdm_pb2.STRING
        elif t == list or t == tuple:
            return rdm_pb2.LIST
        elif t == dict:
            return rdm_pb2.OBJECT
        else:
            raise TypeError(f"wrong value type: {t}")


    @staticmethod
    def map_meta(value: DomainModelMeta) -> rdm_pb2.Signature:
        return rdm_pb2.Signature(
            fqn=value.name, # TODO: change to real fqn
            name=value.name,
            descriptors={
                k: rdm_pb2.SymbolDescriptor(
                    type=rdm_pb2.FUNCTION,
                    name=v.name,
                    arguments=[Py2Proto._map_type(a) for a in v.args],
                    return_value=Py2Proto._map_type(v.ret),
                )
                for k, v in value.domain_functions.items()
            }
        )

    @staticmethod
    def map(value: Union[List[int], List[bool], List[float], List[str], int, float, str, bool, dict]) -> rdm_pb2.SemanticValue:
        if value == None or isinstance(value, bool):
            return Py2Proto._map_bool(value)
        elif isinstance(value, int) or isinstance(value, float):
            return Py2Proto._map_num(value)
        elif isinstance(value, str):
            return Py2Proto._map_string(value)
        elif isinstance(value, list) or isinstance(value, tuple):
            return Py2Proto._map_list(value)
        elif isinstance(value, dict):
            return Py2Proto._map_dict(value)
        else:
            raise TypeError(f"wrong value type: {type(value)}")

